"""Registry of functions for use by ChatCompletions.

Example usage:

    from chatlab import FunctionRegistry
    from pydantic import BaseModel

    registry = FunctionRegistry()

    class Parameters(BaseModel):
        name: str

    from datetime import datetime
    from pytz import timezone, all_timezones, utc
    from typing import Optional
    from pydantic import BaseModel

    def what_time(tz: Optional[str] = None):
        '''Current time, defaulting to the user's current timezone'''
        if tz is None:
            pass
        elif tz in all_timezones:
            tz = timezone(tz)
        else:
            return 'Invalid timezone'
        return datetime.now(tz).strftime('%I:%M %p')

    class WhatTime(BaseModel):
        timezone: Optional[str]

    import chatlab
    registry = chatlab.FunctionRegistry()

    conversation = chatlab.Conversation(
        function_registry=registry,
    )

    conversation.submit("What time is it?")

"""

import inspect
import json
from typing import Callable, Optional, Union, get_args, get_origin

from pydantic import BaseModel

from chatlab.builtins import run_cell


class FunctionArgumentError(Exception):
    """Exception raised when a function is called with invalid arguments."""

    pass


class UnknownFunctionError(Exception):
    """Exception raised when a function is called that is not registered."""

    pass


# Allowed types for auto-inferred schemas
ALLOWED_TYPES = [int, str, bool, float, list, dict]

JSON_SCHEMA_TYPES = {
    int: 'integer',
    float: 'number',
    str: 'string',
    bool: 'boolean',
    list: 'array',
    dict: 'object',
}


def is_optional_type(t):
    """Check if a type is Optional."""
    return get_origin(t) is Union and len(get_args(t)) == 2 and type(None) in get_args(t)


def generate_function_schema(function: Callable, parameters_model: Optional["BaseModel"] = None):
    """Generate a function schema for sending to OpenAI."""
    doc = function.__doc__
    func_name = function.__name__

    if not func_name:
        raise Exception("Function must have a name")
    if func_name == "<lambda>":
        raise Exception("Lambdas cannot be registered. Use `def` instead.")
    if not doc:
        raise Exception("Only functions with docstrings can be registered")

    schema = None
    if parameters_model is not None:
        schema = parameters_model.schema()
    else:
        schema_properties = {}
        sig = inspect.signature(function)
        for name, param in sig.parameters.items():
            if param.annotation == inspect.Parameter.empty:
                raise Exception(f"Parameter {name} of function {func_name} must have a type annotation")

            if is_optional_type(param.annotation):
                actual_type = get_args(param.annotation)[0]

                if actual_type not in ALLOWED_TYPES:
                    raise Exception(
                        f"Type annotation of parameter {name} in function {func_name} "
                        f"must be a JSON serializable type ({ALLOWED_TYPES})"
                    )

                schema_properties[name] = {
                    "type": JSON_SCHEMA_TYPES[actual_type],
                }

            elif param.annotation in ALLOWED_TYPES:
                schema_properties[name] = {
                    "type": JSON_SCHEMA_TYPES[param.annotation],
                }

            else:
                raise Exception(f"Type annotation of parameter {name} in function {func_name} is not allowed")

        schema = {"type": "object", "properties": {}, "required": []}
        if len(schema_properties) > 0:
            schema = {
                "type": "object",
                "properties": schema_properties,
                "required": [
                    name
                    for name, param in sig.parameters.items()
                    if param.default == inspect.Parameter.empty and param.annotation != Optional
                ],
            }

    if schema is None:
        raise Exception(f"Could not generate schema for function {func_name}")

    return {
        "name": func_name,
        "description": doc,
        "parameters": schema,
    }


class FunctionRegistry:
    """Captures a function with schema both for sending to OpenAI and for executing locally."""

    __functions: dict[str, Callable]
    __schemas: dict[str, dict]

    def __init__(self, allow_hallucinated_python: bool = False):
        """Initialize a FunctionRegistry object."""
        self.__functions = {}
        self.__schemas = {}
        self.allow_hallucinated_python = allow_hallucinated_python

    def register(
        self, function: Callable, parameters_model: Optional["BaseModel"] = None, json_schema: Optional[dict] = None
    ):
        """Register a function with a schema for sending to OpenAI."""
        if parameters_model is not None and json_schema is not None:
            raise Exception("Cannot specify both parameters_model and json_schema")

        # Assume that if json_schema is passed, it is valid and complete
        # json_schema takes precedence over parameters_model and auto-inferred schema
        schema = json_schema
        if schema is None:
            schema = generate_function_schema(function, parameters_model)

        self.__functions[function.__name__] = function
        self.__schemas[function.__name__] = schema

        return schema

    def get(self, function_name):
        """Get a function by name."""
        return self.__functions.get(function_name)

    def call(self, name: str, arguments: Optional[str] = None):
        """Call a function by name with the given parameters."""
        function = self.get(name)
        parameters: dict = {}

        # Handle the code interpreter hallucination
        if name == "python" and self.allow_hallucinated_python:
            function = run_cell
            # The "hallucinated" python function takes raw plaintext
            # instead of a JSON object. We can just pass it through.
            parameters = {"code": arguments}
        elif function is None:
            raise UnknownFunctionError(f"Function {name} is not registered")
        elif arguments is None or arguments == "":
            parameters = {}
        else:
            try:
                parameters = json.loads(arguments)
                # TODO: Validate parameters against schema
            except json.JSONDecodeError:
                raise FunctionArgumentError(f"Invalid JSON for parameters of function {name}")

        return function(**parameters)

    def __contains__(self, name):
        """Check if a function is registered by name."""
        return name in self.__functions

    @property
    def function_definitions(self) -> list[dict]:
        """Get a list of function definitions."""
        return list(self.__schemas.values())
